; Graphics routines  -  v2.0

; This include file was re-written by Thomas J. Hruska to work with the
; ASH <-> OShell-82 Development Kit.  Original names were taken from the
; graphics routines of Dines Justsen (developer for ASH).
; The following routines have been added for speed, efficiency, and
; compatibility:
; GOTO_PIXEL       Quicker than FIND_PIXEL in most cases
; GOTO_RS_PIXEL    Same as GOTO_PIXEL except you can use RESTORE_PIXEL define
; FIND_PIXEL       Same as in ROM page 4 without needing to switching pages
; ADJUST_A_HL      Adjusts a and hl according to (b,c)
; ADJUST_HL        Adjusts hl only according to (b,c)
; _GRBUFCPY_V      Same as for ROM_CALL except it works for both shells
; ClrGrBuf         Clears the graph buffer
;
; FIRST SET contains the following functions:
; GetPixel         Tests the pixel at (b,c) and returns the result in a
; Point_on         Turns on the pixel at (b,c)
; Point_off        Turns off the pixel at (b,c)
; Point_chg        Changes the pixel at (b,c)
;
; SECOND SET contains the following functions for use with _GRBUFCPY_V:
; GetPixel_V       Tests the pixel at (b,c) and returns the result in a
; Point_on_V       Turns on the pixel at (b,c)
; Point_off_V      Turns off the pixel at (b,c)
; Point_chg_V      Changes the pixel at (b,c)
;
; SPRITE SET contains the following functions for use with _GRBUFCPY_V:
; SpriteDraw       Draws a (non)clipped sprite at (b,c)
;
; SCROLLING SET contains the following functions for use with _GRBUFCPY_V:
; ScrollUp         Scrolls the graph buffer up and leaves a blank row
; ScrollDown       Scrolls the graph buffer down and leaves a blank row
; ScrollRight      Scrolls the graph buffer right and leaves a blank col
; ScrollLeft       Scrolls the graph buffer left and leaves a blank col
;
; BUFFER SET contains the following routines for use with _GRBUFCPY_V:
; SaveBack         Saves the current Graph Buffer in a secondary buffer
; RestoreBack      Restores the secondary buffer back into the Graph Buffer
; DrawBack         Directly copies the secondary buffer to the LCD
; ClearBack        Clears the secondary buffer out
; GetPixel_Back    Tests the pixel at (b,c) and returns the result in a
; Point_on_Back    Turns on the pixel at (b,c)
; Point_off_Back   Turns off the pixel at (b,c)
; Point_chg_Back   Changes the pixel at (b,c)
; SpriteDraw_B     Draws a (non)clipped sprite at (b,c)
; ScrollUp_B       Scrolls the secondary buffer up and leaves a blank row
; ScrollDown_B     Scrolls the secondary buffer down and leaves a blank row
; ScrollRight_B    Scrolls the secondary buffer right and leaves a blank col
; ScrollLeft_B     Scrolls the secondary buffer left and leaves a blank col
;
; If you translate code that uses reverse graph coordinates, define
; REVERSE_COORD or the result will be upside down :)
; Define FIRST_SET_OFF if you won't be using the first set of functions.
; Define SECOND_SET_OFF if you won't be using the second set of functions.
; Define PIXEL_SET_OFF if you won't be using GOTO_PIXEL, GOTO_RS_PIXEL,
;   FIND_PIXEL, or any of the first set of functions.
; Define FIND_PIXEL_OFF to turn off the FIND_PIXEL routine.
; Define GRBUFCPY_V_OFF if you won't be using _GRBUFCPY_V, ClrGrBuf, any of
;   the second set of functions, SpriteDraw, the scrolling set of functions,
;   or any of the secondary buffer functions.
; Define ADJUST_SET_OFF if you won't be using any of the second set of
;   functions, ADJUST_A_HL, or ADJUST_HL.
; Define SPRITE_SET_OFF if you won't be using SpriteDraw.
; Define SCROLLING_SET_OFF if you won't be using any scrolling functions.
; Define BUFFER_SET_OFF if you won't be using any secondary buffer functions.
; Define MULTIPLE_LAYERS when you plan to use more buffers than 2 (GRAPH_MEM
;   and the secondary buffer).  The secondary buffer functions will
;   automatically be set up to use multiple layers after defining this.
;
; Define ROM_FIND_PIXEL before graph.h to use the FIND_PIXEL routine on ROM
;   page 4.  The single pixel graphics routines will automatically change to
;   use the old FIND_PIXEL routine.
;
; NOTE:  All sets that are turned off affect the Buffer Set.  Also, when
;        using MULTIPLE_LAYERS, you MUST point HL to the buffer you are using
;        before calling any secondary buffer function.
;
; Make sure you also include "graph.h" or you will get a lot of errors.
; Also, make sure you CALL_() the routines (i.e. CALL_(FIND_PIXEL) instead of
; CALL FIND_PIXEL).
;
; Updated routines by Thomas J. Hruska (C) 1997 Shining Light Productions
; Original routines by Dines Justesen

#IFNDEF ROM_FIND_PIXEL
#IFNDEF PIXEL_SET_OFF
#IFNDEF FIRST_SET_OFF
GetPixel:
  push bc                  ; Temporarily store bc on the stack
  CALL_(GOTO_PIXEL)        ; Call the specially designed routine
  ld b,a                   ; a -> b
  READ_PIXEL               ; Reads in the byte into a
  and b                    ; "Tests" the proper pixel in a by ANDing b
  pop bc                   ; Retrieve bc from the stack
  ret                      ; Return to caller

Point_on:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  CALL_(GOTO_RS_PIXEL)     ; Call the specially designed routine
  ld b,a                   ; a -> b
  READ_PIXEL               ; Reads in the byte into a
  or b                     ; Turn "on" the proper pixel in a by ORing b
  RESTORE_PIXEL            ; Restores the positioning of the controller
  WRITE_PIXEL              ; Writes a out to the display controller
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller

Point_off:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  CALL_(GOTO_RS_PIXEL)     ; Call the specially designed routine
  cpl                      ; xor $FF (reverses all bits in a)
  ld b,a                   ; a -> b
  READ_PIXEL               ; Reads in the byte into a
  and b                    ; Turns "off" the proper pixel in a by ANDing b
  RESTORE_PIXEL            ; Restores the positioning of the controller
  WRITE_PIXEL              ; Writes a out to the display controller
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller

Point_chg:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  CALL_(GOTO_RS_PIXEL)     ; Call the specially designed routine
  ld b,a                   ; a -> b
  READ_PIXEL               ; Reads in the byte into a
  xor b                    ; "Changes" the proper pixel in a by XORing b
  RESTORE_PIXEL            ; Restores the positioning of the controller
  WRITE_PIXEL              ; Writes a out to the display controller
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller
#ENDIF


GOTO_PIXEL:                ; DO NOT use with RESTORE_PIXEL (makes it faster
					  ;  if its not needed)
  push bc                  ; Temporarily store bc on the stack
  srl b                    ; Divide b by 8 (there are 8 pixels across per
  srl b                    ; byte)
  srl b
#IFDEF REVERSE_COORD       ; If REVERSE_COORD is defined
  ld a,$3F                 ; $3F -> a  (max. # of lines)
  sub c                    ; a = a - c  (a contains the reversed coordinates)
  ld c,$80                 ; Goto line in display controller
#ELSE
  ld a,$80                 ; Goto line in display controller
#ENDIF
  add a,c                  ; Add $80 + line number to get the true address
  call $7F3                ; Call the display controller delay routine
  out ($10),a              ; Move the controller pointer to the line desired
  ld a,$20                 ; Goto byte in display controller
  add a,b                  ; B is already calculated, add $20 + byte desired
  call $7F3                ; Call the display controller delay routine
  out ($10),a              ; Move the controller pointer to the byte desired
  pop bc                   ; Retrieve bc from the stack
  
  push hl                  ; Temporarily store hl on the stack
  push de                  ; Temporarily store de on the stack
  ld hl,GO_A               ; address of GO_A -> hl
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
  add hl,de                ; hl = hl + de  (absolute address)
  ld a,b                   ; b -> a
  and %00000111            ; Remove the upper 5 bits
  add a,-7                 ; a = a - 7
  neg                      ; make a positive
  add a,a                  ; Multiply by 8
  add a,a
  add a,a
  or $C7                   ; 11???111 = set ?,a
  ld (hl),a                ; a -> (hl)
  ld a,0                   ; 0 -> a
  .db $CB                  ; Op-Code for set b,a  (Part I)
GO_A:
  .db $00                  ; Part II of the Op_Code for set b,a
  pop de                   ; Retrieve de from the stack
  pop hl                   ; Retrieve hl from the stack
  ret                      ; Return to caller


GOTO_RS_PIXEL:             ; Used with RESTORE_PIXEL (usually faster)
  push bc                  ; Temporarily store bc on the stack
  srl b                    ; Divide b by 8 (there are 8 pixels across per
  srl b                    ; byte)
  srl b
#IFDEF REVERSE_COORD       ; If REVERSE_COORD is defined
  ld a,$3F                 ; $3F -> a  (max. # of lines)
  sub c                    ; a = a - c  (a contains the reversed coordinates)
  ld c,$80                 ; Goto line in display controller
#ELSE
  ld a,$80                 ; Goto line in display controller
#ENDIF
  add a,c                  ; Add $80 + line number to get the true address
  ld ($8011),a             ; Store line for a later restore
  call $7F3                ; Call the display controller delay routine
  out ($10),a              ; Move the controller pointer to the line desired
  ld a,$20                 ; Goto byte in display controller
  add a,b                  ; B is already calculated, add $20 + byte desired
  ld ($800F),a             ; Store byte for a later restore
  call $7F3                ; Call the display controller delay routine
  out ($10),a              ; Move the controller pointer to the byte desired
  pop bc                   ; Retrieve bc from the stack
  
  push hl                  ; Temporarily store hl on the stack
  push de                  ; Temporarily store de on the stack
  ld hl,GO_RS_A            ; address of GO_RS_A -> hl
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
  add hl,de                ; hl = hl + de  (absolute address)
  ld a,b                   ; b -> a
  and %00000111            ; Remove the upper 5 bits
  add a,-7                 ; a = a - 7
  neg                      ; make a positive
  add a,a                  ; Multiply by 8
  add a,a
  add a,a
  or $C7                   ; 11???111 = set ?,a
  ld (hl),a                ; a -> (hl)
  ld a,0                   ; 0 -> a
  .db $CB                  ; Op-Code for set b,a  (Part I)
GO_RS_A:
  .db $00                  ; Part II of the Op_Code for set b,a
  pop de                   ; Retrieve de from the stack
  pop hl                   ; Retrieve hl from the stack
  ret                      ; Return to caller


#IFNDEF FIND_PIXEL_OFF
FIND_PIXEL:                ; Used with RESTORE_PIXEL (sometimes faster)
					  ; (This is equivilent to FIND_PIXEL on ROM page 4
					  ;  except it doesn't alter hl)
  push bc                  ; Temporarily store bc on the stack
  srl b                    ; Divide b by 8 (there are 8 pixels across per
  srl b                    ; byte)
  srl b
#IFDEF REVERSE_COORD       ; If REVERSE_COORD is defined
  ld a,$3F                 ; $3F -> a  (max. # of lines)
  sub c                    ; a = a - c  (a contains the reversed coordinates)
  ld c,$80                 ; Goto line in display controller
#ELSE
  ld a,$80                 ; Goto line in display controller
#ENDIF
  add a,c                  ; Add $80 + line number to get the true address
  ld ($8011),a             ; Store line for a later restore
  ld a,$20                 ; Goto byte in display controller
  add a,b                  ; B is already calculated, add $20 + byte desired
  ld ($800F),a             ; Store byte for a later restore
  pop bc                   ; Retrieve bc from the stack
  
  push hl                  ; Temporarily store hl on the stack
  push de                  ; Temporarily store de on the stack
  ld hl,FIND_A             ; address of FIND_A -> hl
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
  add hl,de                ; hl = hl + de  (absolute address)
  ld a,b                   ; b -> a
  and %00000111            ; Remove the upper 5 bits
  add a,-7                 ; a = a - 7
  neg                      ; make a positive
  add a,a                  ; Multiply by 8
  add a,a
  add a,a
  or $C7                   ; 11???111 = set ?,a
  ld (hl),a                ; a -> (hl)
  ld a,0                   ; 0 -> a
  .db $CB                  ; Op-Code for set b,a  (Part I)
FIND_A:
  .db $00                  ; Part II of the Op_Code for set b,a
  pop de                   ; Retrieve de from the stack
  pop hl                   ; Retrieve hl from the stack
  ret                      ; Return to caller
#ENDIF
#ENDIF
#ELSE
#IFNDEF PIXEL_SET_OFF
#IFNDEF FIRST_SET_OFF
GetPixel:
  push bc                  ; Temporarily store bc on the stack
  push hl                  ; Temporarily store hl on the stack
  CALL FIND_PIXEL          ; Call FIND_PIXEL
  RESTORE_PIXEL            ; Restores the positioning of the controller
  ld b,a                   ; a -> b
  READ_PIXEL               ; Reads in the byte into a
  and b                    ; "Tests" the proper pixel in a by ANDing b
  pop hl                   ; Retrieve hl from the stack
  pop bc                   ; Retrieve bc from the stack
  ret                      ; Return to caller

Point_on:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push hl                  ; Temporarily store hl on the stack
  CALL FIND_PIXEL          ; Call FIND_PIXEL
  RESTORE_PIXEL            ; Restores the positioning of the controller
  ld b,a                   ; a -> b
  READ_PIXEL               ; Reads in the byte into a
  or b                     ; Turn "on" the proper pixel in a by ORing b
  RESTORE_PIXEL            ; Restores the positioning of the controller
  WRITE_PIXEL              ; Writes a out to the display controller
  pop hl                   ; Retrieve hl from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller

Point_off:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push hl                  ; Temporarily store hl on the stack
  CALL FIND_PIXEL          ; Call FIND_PIXEL
  RESTORE_PIXEL            ; Restores the positioning of the controller
  cpl                      ; xor $FF (reverses all bits in a)
  ld b,a                   ; a -> b
  READ_PIXEL               ; Reads in the byte into a
  and b                    ; Turns "off" the proper pixel in a by ANDing b
  RESTORE_PIXEL            ; Restores the positioning of the controller
  WRITE_PIXEL              ; Writes a out to the display controller
  pop hl                   ; Retrieve hl from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller

Point_chg:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push hl                  ; Temporarily store hl on the stack
  CALL FIND_PIXEL          ; Call FIND_PIXEL
  RESTORE_PIXEL            ; Restores the positioning of the controller
  ld b,a                   ; a -> b
  READ_PIXEL               ; Reads in the byte into a
  xor b                    ; "Changes" the proper pixel in a by XORing b
  RESTORE_PIXEL            ; Restores the positioning of the controller
  WRITE_PIXEL              ; Writes a out to the display controller
  pop hl                   ; Retrieve hl from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller
#ENDIF
#ENDIF
#ENDIF


#IFNDEF ADJUST_SET_OFF
ADJUST_A_HL:
  push de                  ; Temporarily store de on the stack
  ld hl,AD_A_HL            ; address of AD_A_HL -> hl
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
  add hl,de                ; hl = hl + de  (absolute address)
  ld a,b                   ; b -> a
  and %00000111            ; Remove the upper 5 bits
  add a,-7                 ; a = a - 7
  neg                      ; make a positive
  add a,a                  ; Multiply by 8
  add a,a
  add a,a
  or $C7                   ; 11???111 = set ?,a
  ld (hl),a                ; a -> (hl)
  ld a,0                   ; 0 -> a
  .db $CB                  ; Op-Code for set b,a  (Part I)
AD_A_HL:
  .db $00                  ; Part II of the Op_Code for set b,a

  ld h,0                   ; 0 -> h
  push bc                  ; Temporarily store bc on the stack
#IFDEF REVERSE_COORD       ; If REVERSE_COORD is defined
  push af                  ; Temporarily store af on the stack
  ld a,$3F                 ; $3F -> a  (max. # of lines)
  sub c                    ; a = a - c  (a contains the reversed coordinates)
  ld l,a                   ; a -> l
  pop af                   ; Retrieve af from the stack
#ELSE
  ld l,c                   ; c -> l
#ENDIF
					  ; the following is NOT optimal 7.8% of the time
					  ; multiplies hl by 12
  ld d,h                   ; h -> d
  ld e,l                   ; l -> e
  add hl,hl                ; Multiply hl by 8
  add hl,hl
  add hl,hl
  ex de,hl                 ; Exchange de and hl
  add hl,hl                ; Multiply hl by 4
  add hl,hl
  add hl,de                ; hl = hl + de
  
  srl b
  srl b
  srl b
  ld d,0                   ; 0 -> d
  ld e,b                   ; b -> e
  add hl,de                ; hl = hl + de
  pop bc                   ; Retrieve bc from the stack
  pop de                   ; Retrieve de from the stack
  ret                      ; Return to caller


ADJUST_HL:
  push de                  ; Temporarily store de on the stack
  push bc                  ; Temporarily store bc on the stack
  ld h,0                   ; 0 -> h
#IFDEF REVERSE_COORD       ; If REVERSE_COORD is defined
  push af                  ; Temporarily store af on the stack
  ld a,$3F                 ; $3F -> a  (max. # of lines)
  sub c                    ; a = a - c  (a contains the reversed coordinates)
  ld l,a                   ; a -> l
  pop af                   ; Retrieve af from the stack
#ELSE
  ld l,c                   ; c -> l
#ENDIF
					  ; the following is NOT optimal 7.8% of the time
					  ; multiplies hl by 12
  ld d,h                   ; h -> d
  ld e,l                   ; l -> e
  add hl,hl                ; Multiply hl by 8
  add hl,hl
  add hl,hl
  ex de,hl                 ; Exchange de and hl
  add hl,hl                ; Multiply hl by 4
  add hl,hl
  add hl,de                ; hl = hl + de
  
  pop bc                   ; Retrieve bc from the stack
  ld d,0                   ; 0 -> d
  ld e,b                   ; b -> e
  add hl,de                ; hl = hl + de
  pop de                   ; Retrieve de from the stack
  ret                      ; Return to caller
#ENDIF


#IFNDEF GRBUFCPY_V_OFF
_GRBUFCPY_V:               ; Replaces ROM_CALL(_GRBUFCPY_V)
  push af                  ; Temporarily store all registers on the stack
  push bc
  push de
  push hl
  ld hl,GRAPH_MEM          ; GRAPH_MEM -> hl
_GRBUFCPY_V2:  
  ld a,7                   ; 7 -> a
  call $7F3                ; Call the display controller delay routine
  out ($10),a              ; Change the direction of the pointer after write
  ld a,$80                 ; $80 -> a
LineLoop:
  ld d,a                   ; a -> d
  call $7F3                ; Call the display controller delay routine
  out ($10),a              ; Move the controller pointer to the line desired
  ld a,$20                 ; $20 -> a
  call $7F3                ; Call the display controller delay routine
  out ($10),a              ; Move the controller pointer to the byte desired
  ld bc,$0C11              ; $0C -> b, $11 -> c
WriteLoop:
  call $7F3                ; Call the display controller delay routine
  outi                     ; port (11) (c)=(hl), hl=hl+1, b=b-1
  jr nz,WriteLoop          ; if b<>0 then jump to WriteLoop
  ld a,d                   ; d -> a
  inc a                    ; a = a + 1
  cp $C0                   ; if a = $C0 then return zero
  jr nz,LineLoop           ; if NZ then jump to LineLoop
  pop hl                   ; Retrieve all registers from stack
  pop de
  pop bc
  pop af
  ret                      ; Return to caller


ClrGrBuf:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
  ld hl,GRAPH_MEM          ; GRAPH_MEM -> hl
  ld de,GRAPH_MEM+1        ; GRAPH_MEM+1 -> de
  ld bc,$02FF              ; $2FF -> bc
  ld a,0                   ; 0 -> a
  ld (hl),a                ; a -> (hl)
  ldir                     ; While bc<>0, (de)=(hl), inc hl, dec bc, inc de
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller


#IFNDEF ADJUST_SET_OFF
#IFNDEF SECOND_SET_OFF
GetPixel_V:                ; Tests the pixel at (b,c) in the graph buffer
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
  CALL_(ADJUST_A_HL)       ; Just change a and hl only
  ld b,a                   ; a -> b
  ld de,GRAPH_MEM          ; GRAPH_MEM -> de
  add hl,de                ; hl = hl + de
  ld a,(hl)                ; (hl) -> a
  and b                    ; Test pixel by ANDing b
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  ret                      ; Return to caller


Point_on_V:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
  CALL_(ADJUST_A_HL)       ; Just change a and hl only
  ld b,a                   ; a -> b
  ld de,GRAPH_MEM          ; GRAPH_MEM -> de
  add hl,de                ; hl = hl + de
  ld a,(hl)                ; (hl) -> a
  or b                     ; Turn "on" the pixel by ORing b
  ld (hl),a                ; a -> (hl)
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller


Point_off_V:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
  CALL_(ADJUST_A_HL)       ; Just change a and hl only
  cpl                      ; xor $FF (reverses all bits in a)
  ld b,a                   ; a -> b
  ld de,GRAPH_MEM          ; GRAPH_MEM -> de
  add hl,de                ; hl = hl + de
  ld a,(hl)                ; (hl) -> a
  and b                    ; Turn "off" the pixel by ANDing b
  ld (hl),a                ; a -> (hl)
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller


Point_chg_V:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
  CALL_(ADJUST_A_HL)       ; Just change a and hl only
  ld b,a                   ; a -> b
  ld de,GRAPH_MEM          ; GRAPH_MEM -> de
  add hl,de                ; hl = hl + de
  ld a,(hl)                ; (hl) -> a
  xor b                    ; Changes the pixel by XORing b
  ld (hl),a                ; a -> (hl)
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller
#ENDIF


#IFNDEF SPRITE_SET_OFF
SpriteDraw:                ; Draws a (non)clipped image at (b,c)
  ; This function features clipping of sprites.  Clipping allows you to make
  ; scrolling games such as SQRXZ.  This function is also fast (See the
  ; sprite stats documentation to get average thoroughput for each type of
  ; clipping).
  
  ; This function has a very strict syntax:
  ; hl must be pre-calculated (sprite location + (PROGRAM_ADDR))
  ; (CLIP_X) must be less than b
  ; (CLIP_Y) must be less than c
  ; To avoid crashing your calculator, you also need to make sure you clip
  ; the correct number of pixels off of each sprite

  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  ld de,(CLIP_X)           ; (CLIP_X) -> d, (CLIP_Y) -> e
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
  push ix                  ; Temporarily store ix on the stack
  
  push hl                  ; Temporarily store hl on the stack
  pop ix                   ; Retrieve what was hl into ix  (ix->sprite)
  
  ; Now adjust hl so that it points to where we should begin drawing
  CALL_(ADJUST_A_HL)       ; Adjust a and hl according to memory location
  ld de,GRAPH_MEM          ; GRAPH_MEM -> de
  add hl,de                ; hl = hl + de  (Absolute memory address)
  
SpriteDraw2:  
  ; Get the size of the sprite and transparency code
  push af                  ; Temporarily store af on the stack
  ld b,(ix)                ; (ix) -> b  (X length)
  ld c,(ix+1)              ; (ix) -> c  (Y length)
  
  ld de,(CLIP_X)           ; (CLIP_X) -> d, (CLIP_Y) -> e
  ld a,b                   ; b -> a
  inc d                    ; d = d + 1
  cp d                     ; If a<d then the carry flag will be set
  jr c,Sprite_Error        ; If carry is set, then jump to Sprite_Error
  ld a,c                   ; c -> a
  inc e                    ; e = e + 1
  cp e                     ; If a<e then the carry flag will be set
  jr nc,Sprite_Okay        ; If carry isn't set, then jump to Sprite_Okay
Sprite_Error:
  pop af                   ; Retrieve af from the stack
  pop ix                   ; Retrieve ix from the stack
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  ld (CLIP_X),de           ; d -> (CLIP_X), e -> (CLIP_Y)
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret

Sprite_Okay:
  inc ix                   ; ix = ix + 1
  inc ix                   ; ix = ix + 1
  ld a,(SPRITE_CODE)       ; (SPRITE_CODE) -> a
  and %11111110            ; Reset the transparency bit
  or (ix)                  ; Set/reset the transparency code in a
  ld (SPRITE_CODE),a       ; a -> (SPRITE_CODE)
  inc ix                   ; ix = ix + 1  (start of ix->spritedata)

  bit SPRITE_CLIP_Y_OFF,a  ; Test to see if Y clipping is turned off
  jr nz,Y_Clip_Off         ; If it is, then jump to Y_Clip_Off
  bit SPRITE_CLIP_LOWER,a  ; Test to see if Lower clipping is turned on
  jr nz,Spr_Clip_2         ; If so, then jump to Spr_Clip_2

  ; Clip the upper part of the sprite
  ld a,(CLIP_Y)            ; (CLIP_Y) -> a  (Upper clip)
  or a                     ; Set flags on a
  jr z,Y_Clip_Off          ; If a=0 then jump to Y_Clip_Off
  push bc                  ; Temporarily store bc on the stack
  srl b                    ; Divide b by 8
  srl b
  srl b
  ld d,0                   ; 0 -> d
  ld e,b                   ; b -> e  (de stays the same during this section)
  ld b,a                   ; a -> b
Spr_Clip_Y1:
  add ix,de                ; ix = ix + de
  dec c                    ; c = c - 1
  djnz Spr_Clip_Y1         ; Jump to Spr_Clip_Y1
  pop de                   ; Retrieve what was bc into de
  ld b,d                   ; Restore b to the original value
  jr Y_Clip_Off            ; Jump to Y_Clip_Off

  ; Clip the lower part of the sprite
Spr_Clip_2:
  ld a,(CLIP_Y)            ; (CLIP_Y) -> a  (Lower clip)
  ld d,a                   ; a -> d
  ld a,c                   ; c -> a
  sub d                    ; a = a - d  (EQU c = c - (CLIP_Y))
  ld c,a                   ; a -> c

Y_Clip_Off:                ; (CLIP_Y) can now be destroyed
  ld a,(SPRITE_CODE)       ; (SPRITE_CODE) -> a
  bit SPRITE_CLIP_LEFT,a   ; Test to see if Left clipping is turned off
  ld a,(CLIP_X)            ; (CLIP_X) -> a
  jr z,X_Clip_Off          ; If so, then jump to X_Clip_Off
  ld (CLIP_Y),a            ; a -> (CLIP_Y)
  ; Calculate the exact length of b
  ld d,a                   ; a -> d
  ld a,b                   ; b -> a
  sub d                    ; a = a - d  (EQU b = b - (CLIP_X))
  ld (CLIP_X),a            ; a -> (CLIP_X)
  ; From here on, b and c are correctly set

  ld a,b                   ; b -> a
  cp 8                     ; If a<8, then return carry
  jr c,Small_Sprite        ; If carry, then jump to Small_Sprite
  ld a,(CLIP_Y)            ; (CLIP_Y) -> a
  and %00000111            ; Mask off the upper 5 bits
  add a,-8                 ; a = a - 8
  neg                      ; Makes a positive
  ld (TEMP_1),a            ; a -> (TEMP_1)
  pop af                   ; Retrieve af from the stack
  jr Spr_Drw_Row           ; Jump to Spr_Drw_Row

X_Clip_Off:  
  ; Calculate the exact length of b
  ld d,a                   ; a -> d
  ld a,b                   ; b -> a
  sub d                    ; a = a - d  (EQU b = b - (CLIP_X))
  ld (CLIP_X),a            ; a -> (CLIP_X)
  ; From here on, b and c are correctly set
  
  ; Determine if the sprite is smaller than 8x?
  and %11111000            ; Test the upper 5 bits
  jr z,Small_Sprite        ; If zero, then jump to Small_Sprite
  ld a,8                   ; 8 -> a
  ld (TEMP_1),a            ; a -> (TEMP_1)
  pop af                   ; Retrieve af from the stack
  jr Spr_Drw_Row           ; Jump to Spr_Drw_Row
Small_Sprite:
  ld a,(CLIP_X)            ; (CLIP_X) -> a
  ld (TEMP_1),a            ; a -> (TEMP_1)
  pop af                   ; Retrieve af from the stack

Spr_Drw_Row:
  ; Shift byte according to CLIP_X
  ld e,a                   ; a -> e
  push af                  ; Temporarily store af on the stack
  push hl                  ; Temporarily store hl on the stack
  ld d,(ix)                ; (ix) -> d
  ld a,(CLIP_X)            ; (CLIP_X) -> a
  ld b,a                   ; a -> b
  push bc                  ; Temporarily store bc on the stack
  ld a,(SPRITE_CODE)       ; (SPRITE_CODE) -> a
  bit SPRITE_CLIP_LEFT,a   ; Test to see if Left clipping is turned off
  jr z,Skip_Left_Clip      ; If so, then jump to Skip_Left_Clip
  push bc                  ; Temporarily store bc on the stack
  ld a,(CLIP_Y)            ; (CLIP_Y) -> a
  ld b,0                   ; 0 -> b
  ld c,a                   ; a -> c
  srl c                    ; Divide c by 8
  srl c
  srl c
  add ix,bc                ; ix = ix + bc
  ld d,(ix)                ; (ix) -> d
  and %00000111            ; Get remaining bits to shift
  jr z,Skip_Drw_Row        ; If zero, then jump to Skip_Drw_Row
  ld b,a                   ; a -> b
Spr_Drw_Row_Lp2:
  rl d                     ; Rotate d left
  djnz Spr_Drw_Row_Lp2     ; b = b - 1, if b<>0 then jump to Spr_Drw_Row_Lp2
Skip_Drw_Row:  
  pop bc                   ; Retrieve bc from the stack

Skip_Left_Clip:
  ld a,(TEMP_1)            ; (TEMP_1) -> a
  ld b,a                   ; a -> b
  push bc                  ; Temporarily store bc on the stack
Spr_Drw_Col:
  rl d                     ; Rotate d left
  jr c,Pixel_Is_On         ; If the carry is set, then jump to Pixel_Is_On
  ld a,(SPRITE_CODE)       ; (SPRITE_CODE) -> a
  bit SPRITE_TRANS,a       ; Test to see if transparency is set
  jr nz,Next_Col           ; If the sprite is transparent then jump
Sprite_Not_Trans:
  ld a,e                   ; e -> a
  cpl                      ; XOR $FF  (Reverses all bits)
  and (hl)                 ; Turns "off" the pixel in a by ANDing (hl)
  ld (hl),a                ; a -> (hl)
  jr Next_Col              ; Jump to Next_Col
Pixel_Is_On:  
  ld a,e                   ; e -> a
  or (hl)                  ; Turn "on" the pixel in a by ORing (hl)
  ld (hl),a                ; a -> (hl)
Next_Col:
  rrc e                    ; Rotate e right
  jr nc,Same_Byte          ; If the carry is set, then jump to Same_Byte
  inc hl                   ; hl = hl + 1
Same_Byte:
  djnz Spr_Drw_Col         ; b = b - 1, If b<>0 then jump to Spr_Drw_Col
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve what was bc into af
  inc ix                   ; ix = ix + 1
  sub b                    ; a = a - b
  jr z,Spr_Next_Row        ; If a is zero, then jump to Spr_Next_Row
  ld d,(ix)                ; (ix) -> d
  push af                  ; Temporarily store af on the stack
  and %11111000            ; Test the last 5 bits
  jr z,Last_Byte           ; If zero, then jump to Last_Byte
  ld b,8                   ; 8 -> b
  push bc                  ; Temporarily store bc on the stack
  jr Spr_Drw_Col           ; Jump to Spr_Drw_Col
Last_Byte:
  pop af                   ; Retrieve af from the stack
  push af                  ; Temporarily store af on the stack
  ld b,a                   ; a -> b
  push bc                  ; Temporarily store bc on the stack
  jr Spr_Drw_Col           ; Jump to Spr_Drw_Col
Spr_Next_Row:
  pop hl                   ; Retrieve hl from the stack
  ld de,$000C              ; $000C -> de
  add hl,de                ; hl = hl + de
  pop af                   ; Retrieve af from the stack
  dec c                    ; c = c - 1
  jr nz,Spr_Drw_Row        ; If c is not zero, then jump to Spr_Drw_Row
Sprite_End:
  pop ix                   ; Retrieve ix from the stack
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  ld (CLIP_X),de           ; d -> (CLIP_X), e -> (CLIP_Y)
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller
#ENDIF
#ENDIF


#IFNDEF SCROLLING_SET_OFF
; The 4 scrolling routines scroll the graph buffer in the specified
; direction.  The best (and fastest) way to keep your enemies and main
; character from scrolling, is to remove them beforehand, move them and then
; re-draw them after the scroll.
; LEFT and RIGHT are the slowest methods of scrolling.

ScrollLeft:
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
  ld hl,GRAPH_MEM-$0C      ; GRAPH_MEM-12 -> hl  (Start of scroll)
ScrollLeft2:  
  ld bc,$0B40              ; 11 -> b, 64 -> c
  ld de,$0017              ; 23 -> de  (remains the same throughout scroll)
SclLftLine:
  add hl,de                ; hl = hl + de
  sla (hl)                 ; C <- |7...0| <- 0
SclLftByte:
  dec hl                   ; hl = hl - 1
  rl (hl)                  ; C <- |7...0| <- C
  djnz SclLftByte          ; b = b - 1, if b<>0 then jump
  ld b,$0B                 ; 11 -> b
  dec c                    ; c = c - 1
  jr nz,SclLftLine         ; if not zero, then jump
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  ret                      ; Return to caller

ScrollRight:
  push bc                  ; Temporarily store bc on the stack
  push hl                  ; Temporarily store hl on the stack
  ld hl,GRAPH_MEM-$01      ; GRAPH_MEM-1 -> hl  (Start of scroll)
ScrollRight2:  
  ld bc,$0B40              ; 11 -> b, 64 -> c
SclRtLine:
  inc hl                   ; hl = hl + 1
  srl (hl)                 ; 0 -> |7...0| -> C
SclRtByte:
  inc hl                   ; hl = hl + 1
  rr (hl)                  ; C -> |7...0| -> C
  djnz SclRtByte           ; b = b - 1, if b<>0 then jump
  ld b,$0B                 ; 11 -> b
  dec c                    ; c = c - 1
  jr nz,SclRtLine          ; if not zero, then jump
  pop hl                   ; Retrieve hl from the stack
  pop bc                   ; Retrieve bc from the stack
  ret                      ; Return to caller

ScrollUp:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
  ; First scroll everything
  ld hl,GRAPH_MEM+12       ; GRAPH_MEM+12 -> hl
  ld de,GRAPH_MEM          ; GRAPH_MEM -> de
  ld bc,757                ; 757 -> bc
  ldir                     ; (de)=(hl), hl=hl+1, de=de+1, bc=bc-1
  
  ; Next erase the bottom row
  ld hl,GRAPH_MEM+756      ; GRAPH_MEM+756 -> hl
  ld de,GRAPH_MEM+757      ; GRAPH_MEM+757 -> de
  ld bc,11                 ; 11 -> bc
  ld a,0                   ; 0 -> a
  ld (hl),a                ; a -> (hl)  (EQU 0 -> (GRAPH_MEM))
  ldir                     ; (de)=(hl), hl=hl+1, de=de+1, bc=bc-1
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller

ScrollDown:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
  ; First scroll everything
  ld hl,GRAPH_MEM+756      ; GRAPH_MEM+756 -> hl
  ld de,GRAPH_MEM+768      ; GRAPH_MEM+768 -> de
  ld bc,757                ; 757 -> bc
  lddr                     ; (de)=(hl), hl=hl-1, de=de-1, bc=bc-1
  
  ; Next erase the top row
  ld hl,GRAPH_MEM          ; GRAPH_MEM -> hl
  ld de,GRAPH_MEM+1        ; GRAPH_MEM+1 -> de
  ld bc,11                 ; 11 -> bc
  ld a,0                   ; 0 -> a
  ld (hl),a                ; a -> (hl)  (EQU 0 -> (GRAPH_MEM))
  ldir                     ; (de)=(hl), hl=hl+1, de=de+1, bc=bc-1
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller
#ENDIF


#IFNDEF BUFFER_SET_OFF
; If you use the secondary buffer in any of your programs, they will increase
; in size up to 2K.  This is due to the fact that 768 bytes are being used
; for the secondary buffer.  If you aren't using the secondary buffer, then
; turn the set off to save space.
;
; These routines are mainly used when transparent sprites will pass in front
; of other objects.  This is so that the background behind moving sprites
; isn't destroyed.

SaveBack:
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
#IFNDEF MULTIPLE_LAYERS  
  ld hl,SecondBuffer       ; Address of SecondBuffer -> hl
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
  add hl,de                ; hl = hl + de
#ENDIF
  ex de,hl                 ; de <-> hl
  ld hl,GRAPH_MEM          ; GRAPH_MEM -> hl
  ld bc,$0300              ; 768 -> bc
  ldir                     ; (de)=(hl), hl=hl+1, de=de+1, bc=bc-1
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  ret                      ; Return to caller


RestoreBack:
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
#IFNDEF MULTIPLE_LAYERS  
  ld hl,SecondBuffer       ; Address of SecondBuffer -> hl
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
  add hl,de                ; hl = hl + de
#ENDIF
  ld de,GRAPH_MEM          ; GRAPH_MEM -> de
  ld bc,$0300              ; 768 -> bc
  ldir                     ; (de)=(hl), hl=hl+1, de=de+1, bc=bc-1
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  ret                      ; Return to caller


DrawBack:
  push af                  ; Temporarily store all registers on the stack
  push bc
  push de
  push hl
#IFNDEF MULTIPLE_LAYERS  
  ld hl,SecondBuffer       ; Address of SecondBuffer -> hl
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
  add hl,de                ; hl = hl + de
#ENDIF
  JUMP_(_GRBUFCPY_V2)


ClearBack:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
#IFNDEF MULTIPLE_LAYERS  
  ld hl,SecondBuffer       ; Address of SecondBuffer -> hl
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
  add hl,de                ; hl = hl + de
#ENDIF
  ld d,h                   ; h -> d
  ld e,l                   ; l -> e
  inc de                   ; de = de + 1  (EQU SecondBuffer+1)
  ld bc,$02FF              ; $2FF -> bc
  ld a,0                   ; 0 -> a
  ld (hl),a                ; a -> (hl)
  ldir                     ; While bc<>0, (de)=(hl), inc hl, dec bc, inc de
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller


#IFNDEF ADJUST_SET_OFF
#IFNDEF SECOND_SET_OFF
GetPixel_Back:
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
#IFDEF MULTIPLE_LAYERS  
  ex de,hl                 ; de <-> hl
  CALL_(ADJUST_A_HL)       ; Just change a and hl only
#ELSE
  CALL_(ADJUST_A_HL)       ; Just change a and hl only
  ld de,SecondBuffer       ; SecondBuffer -> de
  add hl,de                ; hl = hl + de
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
#ENDIF
  add hl,de                ; hl = hl + de
  ld b,a                   ; a -> b
  ld a,(hl)                ; (hl) -> a
  and b                    ; Test pixel by ANDing b
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  ret                      ; Return to caller


Point_on_Back:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
#IFDEF MULTIPLE_LAYERS  
  ex de,hl                 ; de <-> hl
  CALL_(ADJUST_A_HL)       ; Just change a and hl only
#ELSE
  CALL_(ADJUST_A_HL)       ; Just change a and hl only
  ld de,SecondBuffer       ; SecondBuffer -> de
  add hl,de                ; hl = hl + de
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
#ENDIF
  add hl,de                ; hl = hl + de
  ld b,a                   ; a -> b
  ld a,(hl)                ; (hl) -> a
  or b                     ; Turn "on" the pixel by ORing b
  ld (hl),a                ; a -> (hl)
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller


Point_off_Back:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
#IFDEF MULTIPLE_LAYERS  
  ex de,hl                 ; de <-> hl
  CALL_(ADJUST_A_HL)       ; Just change a and hl only
#ELSE
  CALL_(ADJUST_A_HL)       ; Just change a and hl only
  ld de,SecondBuffer       ; SecondBuffer -> de
  add hl,de                ; hl = hl + de
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
#ENDIF
  add hl,de                ; hl = hl + de
  cpl                      ; xor $FF (reverses all bits in a)
  ld b,a                   ; a -> b
  ld a,(hl)                ; (hl) -> a
  and b                    ; Turn "off" the pixel by ANDing b
  ld (hl),a                ; a -> (hl)
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller


Point_chg_Back:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
#IFDEF MULTIPLE_LAYERS  
  ex de,hl                 ; de <-> hl
  CALL_(ADJUST_A_HL)       ; Just change a and hl only
#ELSE
  CALL_(ADJUST_A_HL)       ; Just change a and hl only
  ld de,SecondBuffer       ; SecondBuffer -> de
  add hl,de                ; hl = hl + de
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
#ENDIF
  add hl,de                ; hl = hl + de
  ld b,a                   ; a -> b
  ld a,(hl)                ; (hl) -> a
  xor b                    ; Changes the pixel by XORing b
  ld (hl),a                ; a -> (hl)
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller
#ENDIF


#IFNDEF SPRITE_SET_OFF
SpriteDraw_B:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  ld de,(CLIP_X)           ; (CLIP_X) -> d, (CLIP_Y) -> e
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
  push ix                  ; Temporarily store ix on the stack
  
  push hl                  ; Temporarily store hl on the stack
  pop ix                   ; Retrieve what was hl into ix  (ix->sprite)
  
  ; Now adjust hl so that it points to where we should begin drawing
#IFDEF MULTIPLE_LAYERS  
  ex de,hl                 ; de <-> hl
  CALL_(ADJUST_A_HL)       ; Just change a and hl only
#ELSE
  CALL_(ADJUST_A_HL)       ; Just change a and hl only
  ld de,SecondBuffer       ; SecondBuffer -> de
  add hl,de                ; hl = hl + de
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
#ENDIF
  add hl,de                ; hl = hl + de
  JUMP_(SpriteDraw2)
#ENDIF
#ENDIF


#IFNDEF SCROLLING_SET_OFF
; The 4 scrolling routines scroll the secondary buffer in the specified
; direction.  The best (and fastest) way to keep your enemies and main
; character from scrolling, is to remove them beforehand, move them and then
; re-draw them after the scroll.
; LEFT and RIGHT are the slowest methods of scrolling.

ScrollLeft_B:
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
#IFDEF MULTIPLE_LAYERS  
  ld de,-12                ; -12 -> de
#ELSE
  ld hl,SecondBuffer-$0C   ; SecondBuffer-12 -> hl  (Start of scroll)
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
#ENDIF
  add hl,de                ; hl = hl + de
  JUMP_(ScrollLeft2)

ScrollRight_B:
  push bc                  ; Temporarily store bc on the stack
  push hl                  ; Temporarily store hl on the stack
#IFDEF MULTIPLE_LAYERS  
  dec hl                   ; hl = hl - 1
#ELSE
  ld hl,SecondBuffer-$01   ; SecondBuffer-1 -> hl  (Start of scroll)
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
  add hl,de                ; hl = hl + de
#ENDIF
  JUMP_(ScrollRight2)

ScrollUp_B:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
  ; First scroll everything
#IFNDEF MULTIPLE_LAYERS  
  ld hl,SecondBuffer       ; SecondBuffer -> hl
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
  add hl,de                ; hl = hl + de
#ENDIF
  ld d,h                   ; h -> d
  ld e,l                   ; l -> e
  ld bc,12                 ; 12 -> bc
  add hl,bc                ; hl = hl + bc
  ld bc,757                ; 757 -> bc
  ldir                     ; (de)=(hl), hl=hl+1, de=de+1, bc=bc-1
  
  ; Next erase the bottom row
#IFDEF MULTIPLE_LAYERS
  pop hl                   ; Retrieve hl from the stack
  push hl                  ; Temporarily store hl on the stack
  ld bc,756                ; 756 -> bc
  add hl,bc                ; hl = hl + bc
#ELSE  
  ld hl,SecondBuffer+756   ; SecondBuffer+756 -> hl
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
  add hl,de                ; hl = hl + de
#ENDIF  
  ld d,h                   ; h -> d
  ld e,l                   ; l -> e
  inc de                   ; de = de + 1  (EQU SecondBuffer+757)
  ld bc,11                 ; 11 -> bc
  ld a,0                   ; 0 -> a
  ld (hl),a                ; a -> (hl)  (EQU 0 -> (GRAPH_MEM))
  ldir                     ; (de)=(hl), hl=hl+1, de=de+1, bc=bc-1
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller

ScrollDown_B:
  push af                  ; Temporarily store af on the stack
  push bc                  ; Temporarily store bc on the stack
  push de                  ; Temporarily store de on the stack
  push hl                  ; Temporarily store hl on the stack
  ; First scroll everything
#IFDEF MULTIPLE_LAYERS  
  ld bc,768                ; 768 -> bc
  add hl,bc                ; hl = hl + bc
#ELSE
  ld hl,SecondBuffer+768   ; SecondBuffer+768 -> hl
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
  add hl,de                ; hl = hl + de
#ENDIF
  ld d,h                   ; h -> d
  ld e,l                   ; l -> e
  ld bc,-12                ; -12 -> bc
  add hl,bc                ; hl = hl + bc
  ld bc,757                ; 757 -> bc
  lddr                     ; (de)=(hl), hl=hl-1, de=de-1, bc=bc-1
  
  ; Next erase the top row
#IFDEF MULTIPLE_LAYERS
  pop hl                   ; Retrieve hl from the stack
  push hl                  ; Temporarily store hl on the stack
#ELSE  
  ld hl,SecondBuffer       ; SecondBuffer -> hl
  ld de,(PROGRAM_ADDR)     ; (PROGRAM_ADDR) -> de
  add hl,de                ; hl = hl + de
#ENDIF  
  ld d,h                   ; h -> d
  ld e,l                   ; l -> e
  inc de                   ; de = de + 1  (EQU SecondBuffer+757)
  ld bc,11                 ; 11 -> bc
  ld a,0                   ; 0 -> a
  ld (hl),a                ; a -> (hl)  (EQU 0 -> (GRAPH_MEM))
  ldir                     ; (de)=(hl), hl=hl+1, de=de+1, bc=bc-1
  pop hl                   ; Retrieve hl from the stack
  pop de                   ; Retrieve de from the stack
  pop bc                   ; Retrieve bc from the stack
  pop af                   ; Retrieve af from the stack
  ret                      ; Return to caller
#ENDIF


SecondBuffer:
.ORG $+$0300               ; Reserve 768 bytes of space for the buffer

#ENDIF
#ENDIF
